def read_asc(ascii_file):

    import numpy as np
    from linecache import getline

    source1 = ascii_file
    # Parse the header using a loop and
    # the built-in linecache module
    hdr = [getline(source1, i) for i in range(1,7)]
    values = [float(h.strip('\n').strip().split(" ")[-1]) \
     for h in hdr]
    cols,rows,xll,yll,cell,nd = values
    cols = int(cols)
    rows = int(rows)
    delta_x = cell
    delta_y = cell

    values = [(h.split(" ")[0]) for h in hdr]
    s1,s2,s3,s4,s5,s6 = values

    if ( s3=='xllcorner'):

        x_min = xll+0.5*delta_x
        y_min = yll+0.5*delta_y
    
    elif ( s3=='xllcenter'):
    
        x_min = xll
        y_min = yll

    x_max = x_min+(cols-1)*delta_x
    x = np.linspace(x_min,x_max,cols)

    y_max = y_min+(rows-1)*delta_y
    y = np.linspace(y_min,y_max,rows)

    X,Y = np.meshgrid(x,y)

    # Load the dem into a numpy array
    h = np.flipud(np.loadtxt(source1, skiprows=6))

    h[h==nd]=np.nan

    return (X, Y, h, x_min, x_max, delta_x, y_min, y_max, delta_y)





def advance_time( h_init , h_old , delta_t , delta_x , delta_y , lambda_wb , k_wb ,
                  S_c , enne , k , max_nlc , max_inner_iter , res ,
                  Dirichlet, Transient , Neumann , grow_rate , A_c ,
                  simtime , verbose_level ):

    # ADVANCE_TIME This function update of one time step delta_t the solution
    # of dh/dt = div( f(h) grad(h) )
    #
    # where
    #
    # f(h) = k * ( 1 + ( max_nlc -1 ) * ( |grad(h)| / S_c )^n )
    #
    # k is the diffusivity constant, S_c is the tangent of the critical slope.
    # and max_nlc is the maximum value of the non-linear coefficient that
    # multiply k.

    import numpy as np
    from nonlinear_function import nonlinear_function
    from numerical_fluxes import numerical_fluxes
    from relative_coeff import relative_coeff

    nx = h_init.shape[0]
    ny = h_init.shape[1]

    h_temp = np.zeros((nx,ny))

    h_temp[0:nx,0:ny] = h_init[0:nx,0:ny]
    h_new = np.zeros((nx,ny))

    w = np.zeros((nx,ny)) # temporary solution array
    v = np.zeros((nx,ny)) # temporary solution array

    ax = np.zeros((nx)) # lower-diagonal
    bx = np.zeros((nx)) # diagonal
    cx = np.zeros((nx)) # upper-diagonal
    dx = np.zeros((nx)) # right-hand side

    ay = np.zeros((ny)) # lower-diagonal
    by = np.zeros((ny)) # diagonal
    cy = np.zeros((ny)) # upper-diagonal
    dy = np.zeros((ny)) # right-hand side

    expl_term_temp_x = np.zeros((nx,ny))
    expl_term_temp_y = np.zeros((nx,ny))

    expl_term_old_x = np.zeros((nx,ny))
    expl_term_old_y = np.zeros((nx,ny))

    k_rel = relative_coeff(simtime,lambda_wb,k_wb)
    f_old = nonlinear_function(h_old,delta_x,delta_y,S_c,enne,k*k_rel,max_nlc)

    for inner_iter in range(0,max_inner_iter-1):

        if verbose_level == 2:

            print("Outer loop: dt = ",delta_t)

        elif verbose_level > 2:

            print("Outer loop: dt = ",delta_t)

        f = nonlinear_function(h_temp,delta_x,delta_y,S_c,enne,k*k_rel,max_nlc)

        # Evaluate the explicit term (the right-hand side of the linear system
        # for the diffusion in the x-direction)

        flux_temp_east,flux_temp_west,flux_temp_north,flux_temp_south = \
                            numerical_fluxes( h_temp , f , delta_x , delta_y )

        expl_term_temp_x[0,:] = flux_temp_east[0,:] / delta_x

        expl_term_temp_x[1:nx-1,:] = ( flux_temp_east[1:nx-1,:] \
                            - flux_temp_west[1:nx-1,:] ) / delta_x

        expl_term_temp_x[nx-1,:] = - flux_temp_west[nx-1,:] / delta_x

        expl_term_temp_y[0:nx,0] = flux_temp_north[0:nx,0] / delta_y

        expl_term_temp_y[0:nx,1:ny-1] = ( flux_temp_north[0:nx,1:ny-1] \
                            - flux_temp_south[0:nx,1:ny-1] ) / delta_y

        expl_term_temp_y[0:nx,ny-1] = - flux_temp_south[0:nx,ny-1] / delta_y

        #-----------------------------

        flux_old_east,flux_old_west,flux_old_north,flux_old_south = \
                            numerical_fluxes( h_old , f_old , delta_x , delta_y )

        expl_term_old_x[0,:] = flux_old_east[0,:] / delta_x

        expl_term_old_x[1:nx-1,:] = ( flux_old_east[1:nx-1,:] \
                            - flux_old_west[1:nx-1,:] ) / delta_x

        expl_term_old_x[nx-1,:] = - flux_old_west[nx-1,:] / delta_x

        expl_term_old_y[0:nx,0] = flux_old_north[0:nx,0] / delta_y

        expl_term_old_y[0:nx,1:ny-1] = ( flux_old_north[0:nx,1:ny-1] \
                            - flux_old_south[0:nx,1:ny-1] ) / delta_y

        expl_term_old_y[0:nx,ny-1] = - flux_old_south[0:nx,ny-1] / delta_y


        expl_term = - ( h_temp - h_old ) \
                    + delta_t * 0.5 * ( expl_term_temp_x + expl_term_temp_y ) \
                    + delta_t * 0.5 * ( expl_term_old_x + expl_term_old_y )   \
                    + A_c * delta_t


        # Solve the linearized system in the x-direction with the Thomas
        # algorithm.


        if ( Dirichlet[0] ) or ( Transient[0] ):

            bx[0] = 1.0;   # these coefficients are defined in order to impose the
            cx[0] = 0.0;   # Dirichlet boundary condition
            dx[0] = 0.0;   # boundary condition


        if ( Dirichlet[1] ) or ( Transient[1] ):

            first_row = 1

        else:

            first_row = 0

        if ( Dirichlet[2] ) or ( Transient[2] ):

            ax[nx-1] = 0.0   # these coefficients are defined in order to impose
            bx[nx-1] = 1.0   # the Dirichlet boundary condition
            dx[nx-1] = 0.0   # boundary condition

        if ( Dirichlet[3] ) or ( Transient[3] ):

            last_row = ny-1

        else:

            last_row = ny


        for j in range(first_row,last_row):

            if ( Neumann[0] ):

                # coefficients for the Neumann boundary condition at node (0,j)

                bx[0] = 1.0 + 0.25 * delta_t / delta_x**2 * \
                        ( f[0,j] + f[1,j] )

                cx[0] = - 0.25 * delta_t / delta_x**2 * ( f[1,j] + f[0,j] )
                dx[0] = expl_term[0,j]

            ax[1:nx-1] = - 0.25 * delta_t / delta_x**2 * ( f[0:nx-2,j] + f[1:nx-1,j] )

            bx[1:nx-1] = 1.0 + 0.25 * delta_t / delta_x**2 * \
               ( f[2:nx,j] + 2.0 * f[1:nx-1,j] + f[0:nx-2,j] )

            cx[1:nx-1] = - 0.25 * delta_t / delta_x**2 * ( f[2:nx,j] + f[1:nx-1,j] )

            dx[1:nx-1] = expl_term[1:nx-1,j]

            if ( Neumann[2] ):

               # coefficients for the Neumann boundary condition at node (nx,j)
               ax[nx-1] = - 0.25 * delta_t / delta_x**2 * ( f[nx-2,j] + f[nx-1,j] )

               bx[nx-1] = 1.0 + 0.25 * delta_t / delta_x**2 * \
                        ( f[nx-1,j] + f[nx-2,j] )

               dx[nx-1] = expl_term[nx-1,j]

            # call the tridiagonal solver
            # w[0:nx,j] = TDMAsolver(ax,bx,cx,dx)
            w[0:nx,j] = NEWsolver(ax,bx,cx,dx)


        # Solve in the y-direction

        if ( Dirichlet[0] ):

            first_column = 1
            h_temp[0,0:ny] = h_old[0,0:ny]

        elif ( Transient[0] ):

            first_column = 1
            h_temp[0,0:ny] = h_old[0,0:ny] + grow_rate[0]*delta_t

        elif ( Neumann[0] ):

            first_column = 0


        if ( Dirichlet[1] ):

            by[0] = 1.0
            cy[0] = 0.0
            dy[0] = 0.0

        elif ( Transient[1] ):

            by[0] = 1.0
            cy[0] = 0.0
            dy[0] = grow_rate(2)*delta_t

            h_temp[0:nx,0] = h_old[0:nx,0] + grow_rate[1]*delta_t


        if ( Dirichlet[2] ):

            last_column = nx-1
            h_temp[nx-1,0:ny] = h_old[nx-1,0:ny]

        elif ( Transient[2] ):

            last_column = nx-1
            h_temp[nx-1,0:ny] = h_old[nx-1,0:ny] + grow_rate[2]*delta_t

        elif ( Neumann[2] ):

            last_column = nx


        if ( Dirichlet[3] ):

            ay[ny-1] = 0.0
            by[ny-1] = 1.0
            dy[ny-1] = 0.0

        elif ( Transient[3] ):

            ay[ny-1] = 0.0
            by[ny-1] = 1.0
            dy[ny-1] = grow_rate[3]*delta_t

            h_temp[0:nx,ny-1] = h_old[0:nx,ny-1] + grow_rate[3]*delta_t;


        for i in range(first_column,last_column):

            if ( Neumann[1] ):

                # coefficients for the Neumann boundary condition at node [i,0]

                by[0] = 1.0 + 0.25 * delta_t / delta_y**2 * \
                        ( f[i,0] + f[i,1] )
                cy[0] = - 0.25 * delta_t / delta_y**2 * ( f[i,1] + f[i,0] )
                dy[0] = w[i,0]


            ay[1:ny-1] = - 0.25 * delta_t / delta_y**2 * ( f[i,0:ny-2] + f[i,1:ny-1] )
            by[1:ny-1] = 1.0 + 0.25 * delta_t / delta_y**2 * \
                         ( f[i,2:ny] + 2.0 * f[i,1:ny-1] + f[i,0:ny-2] )

            cy[1:ny-1] = - 0.25 * delta_t / delta_y**2 * ( f[i,2:ny] + f[i,1:ny-1] )

            dy[1:ny-1] = w[i,1:ny-1]


            if ( Neumann[3] ):

                # coefficients for the Neumann boundary condition at node [i,ny-1
                ay[ny-1] = - 0.25 * delta_t / delta_y**2 * ( f[i,ny-2] + f[i,ny-1] )
                by[ny-1] = 1.0 + 0.25 * delta_t / delta_y**2 * \
                           ( f[i,ny-1] + f[i,ny-2] )

                dy[ny-1] = w[i,ny-1]

            # call the tridiagonal solver
            # v[i,0:ny] = TDMAsolver(ay,by,cy,dy)
            v[i,0:ny] = NEWsolver(ay,by,cy,dy)

        residual = np.maximum(np.abs(v.min()), np.abs(v.max()) )

        if ( verbose_level >= 3 ):

            print('Inner loop. Iter = ' + str(i) + ' residual = ' + str(residual))


        h_temp[first_column:last_column,first_row:last_row] +=   \
                    v[first_column:last_column,first_row:last_row]

        if ( residual < res ):

            break


    h_new[0:nx,0:ny] = 1.0 * h_temp[0:nx,0:ny]


    return [h_new[0:nx,0:ny],residual]

def NEWsolver(a,b,c,d):

    import numpy as np
    import scipy.linalg as la

    # Create arrays and set values
    m = b.size
    ab = np.zeros((3,m))
    ab[0] = a
    ab[1] = b
    ab[2] = c

    return la.solve_banded ((1,1),ab,d)

def TDMAsolver(a,b,c,d):

    # TDMAsolver is tridiagonal matrix solver using the Thomas algorithm
    # (named after Llewellyn Thomas). It is a simplified form of Gaussian
    # elimination that can be used to solve tridiagonal systems of equations.
    # A tridiagonal system for n unknowns may be written as
    #
    #    a_i*x_{i-1} + b_i*x_i + c_i*x_{i+1} = d_i ,
    #
    # where a a_1=0 and c_n=0.
    # a, b, c, and d are the column 1xn arrays for the compressed tridiagonal
    # matrix.

    import numpy as np

    n = b.size  # n is the number of rows

    # Modify the first-row coefficients
    c[0] /= b[0]      # Division by zero risk.
    d[0] /= b[0]      # Division by zero would imply a singular matrix.

    x = np.zeros((n))

    for i in range(1,n):

        id = 1.0 / (b[i] - c[i-1] * a[i])  # Division by zero risk.
        c[i] *= id                         # Last value calculated is redundant.
        d[i] = (d[i] - d[i-1] * a[i]) * id

    # Now back substitute.
    x[n-1] = d[n-1]

    for i in reversed(range(n-1)):

        x[i] = d[i] - c[i] * x[i + 1]

    return x



def relative_coeff(t,lambda_wb,k_wb):

    # RELATIVE_COEFF This function compute a non-dimensional relative
    # coefficent (0<rel_coeff<=1) multiplying the diffusion coefficient
    # k. It is based on a cumulative Weibull distribution with parameters
    # lambda_wb and k_wb

    import numpy as np

    if (lambda_wb == 0.0 ):
    
        k_rel = 1.0
        
    else:

        k_rel= np.maximum(0.001,1.0-np.exp(-(t/lambda_wb)**k_wb))

    return k_rel



def numerical_fluxes( h , f , delta_x , delta_y ):

    # NUMERICAL_FLUXES This function avaluates the fluxes at the boundaries
    # of each cell (east,west,south,north) with a finite-difference scheme.

    import numpy as np
 
    nx = h.shape[0]
    ny = h.shape[1]


    flux_east = np.zeros((nx,ny))
    flux_west = np.zeros((nx,ny))
    flux_north = np.zeros((nx,ny))
    flux_south = np.zeros((nx,ny))


    flux_west[1:nx,:] = 0.5 * ( f[0:nx-1,:] + f[1:nx,:] ) * \
    ( h[1:nx,:] - h[0:nx-1,:] ) / delta_x

    flux_east[0:nx-1,:] = flux_west[1:nx,:]

    flux_south[:,1:ny] = 0.5 * ( f[:,0:ny-1] + f[:,1:ny] ) * \
    ( h[:,1:ny] - h[:,0:ny-1] ) / delta_y

    flux_north[:,0:ny-1] = flux_south[:,1:ny]

    return (flux_east,flux_west,flux_north,flux_south)




def nonlinear_function(h,delta_x,delta_y,S_c,enne,k,max_nlc):

    # NONLINEAR_FUNCTION This function evaluate the nonlinear function
    #
    # f(h) = k * ( 1 + ( max_nlc -1 ) * ( |grad(h)| / S_c )^n )
    #
    # k is the diffusivity constant, S_c is the tangent of the critical slope.
    # and max_nlc is the maximum value of the non-linear coefficient that
    # multiply k.

    import numpy as np

    nx = h.shape[0]
    ny = h.shape[1]

    h_x,h_y = np.gradient(h,delta_x,delta_y)

    grad_h = np.sqrt( h_x**2 + h_y**2 )

    ratio = np.minimum(np.ones((nx,ny)),grad_h/S_c) #gradient computed in each sells, so wnp.ones is an array

    if ( np.isinf(enne) ):

        non_linear_coeff = 1.0

    else:

        if ( enne == 0 ): #when n=0 we have linear flux(diffusion). enne = n!!!

            non_linear_coeff = max_nlc

        else:

            ratio2 = ( ( max_nlc - 1.0 ) / max_nlc ) ** ( 1.0 / enne ) * ratio #max_nlc=cmax
            non_linear_coeff = 1.0 / ( 1.0 - ratio2**enne )

    effe = k * non_linear_coeff

    return effe



def ErosioNL( X, Y, h, final_time, delta_t_max,delta_t0, cr_angle, \
    enne, k, lambda_wb, k_wb, max_nlc, max_inner_iter, res , bc, grow_rate ,        \
    run_name , n_output , A_c , verbose_level, save_output_flag,   \
    plot_output_flag, plot_show_flag):


    # EROSIONL This function solve for the nonlinear diffusion problem
    #
    # dh/dt = div( f(h) grad(h) )
    #
    # where
    #
    # f(h) = k * ( 1 + ( max_nlc -1 ) * ( |grad(h)| / S_c )^n )
    #
    # k is the diffusivity constant, S_c is the tangent of the critical slope.
    # and max_nlc is the maximum value of the non-linear coefficient that
    # multiply k.
    #
    # - X and Y are the (nx,ny) array of the coordinates as generated by the
    #   matlab meshgrid command;
    # - h is the (nx,ny) initial elevation array;
    # - final_time is the duration of the run;
    # - delta_t0 is the maximum time step;
    # - angle is the critical slope;
    # - enne is the exponent of the model;
    # - k is the diffusion goefficient, a scalar or a (nx,ny) array. The
    #   value of k is multiplied the a relative coefficient k_rel, which
    #   is a function of time;
    # - lambda_wb is the time scale parameter of the Weibull cumulative 
    #   distribution (for t=lambda_wb, k_rel=(e-1)/e). If lambda_wb=0,
    #   then k_rel is set to 1.0;  
    # - k_wb is the shape parameter of the Weibull distribution;
    # - max_nlc is the maximum value of the nonlienar coefficient;
    # - max_inner_iter is the maximum number of iteration for the inner loop;
    # - res is the residual required for the convergenge of the inner loop;
    # - bc is a 1x4 array with the boundary conditions at the S,W,N and E
    #   boundaries and the value can be: 'N' (Neumann), 'D' (Dirichlet) or
    #   'T' transient.
    # - grow_rate is a 1x4 array with the rates of change at the boundaries
    #   when the 'transient' boundary condition is applied;
    # - run_name is the name given to the output files
    # - n_output is the number of output plotted on screen.
    #
    # The model is a modification of the model presented in:
    #   Pelletier, J. # Cline, M., Nonlinear slope-dependent sediment transport
    #   in cinder cone evolution, Geology, Geological Society of America, 2007,
    #   35, 1067.
    # The numerical scheme adopted is described in:
    #   Tian, Z. & Ge, Y., A fourth-order compact ADI method for solving
    #   two-dimensional unsteady convection-diffusion problems, Journal of
    #   computational and applied mathematics, Elsevier, 2007, 198, 268-286.
    #
    # See also GRADIENT

    import numpy as np


    from advance_time import advance_time
    from nonlinear_function import nonlinear_function
    from make_plot import make_plot
    import netCDF4
    import time
    import sys

    nx = h.shape[0]
    ny = h.shape[1]

    h_old = np.zeros((nx,ny))
    h_new = np.zeros((nx,ny))
    h_temp = np.zeros((nx,ny))
    h_temp_half = np.zeros((nx,ny))
    h_temp_full = np.zeros((nx,ny))

    if ( np.isscalar(k) ):

        k = k * np.ones((nx,ny))

    simtime = 0.0
    iter = 0

    iteration_draw_times = np.linspace(final_time/n_output,final_time,n_output)

    h_init = h

    h_min = np.amin(h) - 0.1*(np.amax(h)-np.amin(h))
    h_max = np.amax(h) + 0.1*(np.amax(h)-np.amin(h))

    if ( plot_output_flag ):

        make_plot(X,Y,h,h_init,h_min,h_max,simtime,cr_angle,run_name,iter,plot_show_flag)

    if ( save_output_flag):

        # Save initial topography on ascii raster file
        header = "ncols     %s\n" % h.shape[1]
        header += "nrows    %s\n" % h.shape[0]
        header += "xllcenter " + str(np.amin(X)) +"\n"
        header += "yllcenter " + str(np.amin(Y)) +"\n"
        header += "cellsize " + str(np.abs(X[1,2]-X[1,1])) +"\n"
        header += "NODATA_value -9999\n"

        output_full = run_name + '_{0:03}'.format(0) + '.asc'

        np.savetxt(output_full, np.flipud(h), header=header, fmt='%1.5f',comments='')


    # create netcdf4 file
    ncfilename = run_name+'.nc'

    ncfile = netCDF4.Dataset(ncfilename,mode='w',format='NETCDF4')

    x_dim = ncfile.createDimension('x', nx)
    y_dim = ncfile.createDimension('y', ny)
    time_dim = ncfile.createDimension('time', None) # unlimited axis (can be appended to).

    ncfile.title = run_name+' output'

    ncfile.Conventions = "CF-1.0"
    ncfile.subtitle="My model data subtitle"
    ncfile.anything="write anything"


    x = ncfile.createVariable('x', np.float64, ('x',),zlib=True)
    x.long_name = 'x dim'
    x.units = 'meters'

    y = ncfile.createVariable('y', np.float64, ('y',),zlib=True)
    y.long_name = 'y dim'
    y.units = 'meters'

    t = ncfile.createVariable('time', np.float64, ('time',))
    t.long_name = 'Time'
    t.units = 'seconds'

    z = ncfile.createVariable('z',np.float64,('time','y','x'),zlib=True) # note: unlimited dimension is leftmost
    z.standard_name = 'flow thickness' # this is a CF standard name
    z.units = 'meters'

    dz_dt = ncfile.createVariable('dz_dt',np.float64,('time','y','x'),zlib=True) # note: unlimited dimension is leftmost
    dz_dt.standard_name = 'erosion/deposit rate' # this is a CF standard name
    dz_dt.units = 'meters/seconds'

    dz_tot = ncfile.createVariable('dz_tot',np.float64,('time','y','x'),zlib=True) # note: unlimited dimension is leftmost
    dz_tot.standard_name = 'flow erosion and deposit thickness' # this is a CF standard name
    dz_tot.units = 'meters'

    t[iter] = simtime
    z[iter,:,:] = h
    dz_dt[iter,:,:] = h - h_init
    dz_tot[iter,:,:] = h - h_init

    x[:] = X[0,:]
    y[:] = Y[:,0]

    delta_x = np.abs(X[1,2]-X[1,1])
    delta_y = np.abs(Y[2,1]-Y[1,1])

    h_x,h_y = np.gradient(h,delta_x,delta_y)
    grad_h = np.sqrt( h_x**2 + h_y**2 )

    max_angle = np.zeros((n_output))
    mean_angle = np.zeros((n_output))
    time_iter = np.zeros((n_output))


    max_angle[iter] = np.arctan(np.max(np.max(grad_h)))*180/np.pi
    mean_angle[iter] = np.arctan(np.mean(grad_h))*180/np.pi

    time_iter[iter] = simtime

    print('Maximum/mean slope of the DEM = ' + str(max_angle[iter]))

    print('Mean slope of the DEM = ' + str(mean_angle[iter]))

    if ( cr_angle < max_angle[iter] ):

        print('Critical slope = ' + str(cr_angle))
        print('WARNING: critical slope < maximum slope of the DEM')


    S_c = np.tan(cr_angle/180.0*np.pi)

    # set the variables for the boundary conditions

    Neumann = np.zeros((4), dtype=bool)
    Dirichlet = np.zeros((4), dtype=bool)
    Transient = np.zeros((4), dtype=bool)


    for i in range(0,4):

        if ( bc[i] == 'N' ):

            Neumann[i] = True
            Dirichlet[i] = False
            Transient[i] = False

        elif ( bc[i] == 'D' ):

            Dirichlet[i] = True
            Neumann[i] = False
            Transient[i] = False

        elif ( bc[i] == 'T' ):

            Transient[i] = True
            Dirichlet[i] = False
            Neumann[i] = False

        else:

            print('WARNING: wrong boundary condition, set to Neumann')
            Dirichlet[0:3] = False
            Neumann[0:3] = True




    f = nonlinear_function(h,delta_x,delta_y,S_c,enne,k,max_nlc)

    delta_t_stab = 0.5 * delta_x * delta_y / np.max(f)

    delta_t = np.minimum( delta_t0 , delta_t_stab )
    delta_t_old = delta_t

    print('Delta t for stability = ', str(delta_t_stab))


    err = 1.0
    err_old = 1.0

    if sys.version_info >= (3, 0):

        start = time.process_time()

    else:

        start = time.clock()


    while ( simtime < final_time):

        delta_t_orig = 0.5 * ( delta_t + delta_t_old )

        coeff_t = np.maximum( 0.85 , np.minimum( 1.05,err**( -0.35 ) * err_old**( 0.2 ) ) )

        delta_t_old = delta_t
        err_old = err

        delta_t_orig = delta_t_orig * coeff_t
        delta_t_orig = min(delta_t_orig,iteration_draw_times[iter]-simtime)

        err = 1.0

        h_old[0:nx,0:ny] = h[0:nx,0:ny]

        while ( err >= 1.0 ):

            # integration with two steps with half step-size

            delta_t = 0.5 * delta_t_orig

            # first half step


            (h_temp[0:nx,0:ny],residual) = \
                    advance_time( h_old[0:nx,0:ny] , h_old[0:nx,0:ny] , delta_t , delta_x , \
                    delta_y ,  lambda_wb , k_wb , S_c , enne , k , max_nlc , max_inner_iter , res , \
                    Dirichlet, Transient , Neumann , grow_rate , A_c , \
                    simtime, verbose_level )

            # print '1st half ',np.min(h_temp),np.max(h_temp)

            # second half step

            h_guess = h_temp + ( h_temp - h_old)

            (h_new[0:nx,0:ny],residual) = \
                    advance_time( h_guess[0:nx,0:ny] , h_temp[0:nx,0:ny] , delta_t , delta_x , \
                    delta_y ,  lambda_wb , k_wb , S_c , enne , k , max_nlc , max_inner_iter , res , \
                    Dirichlet, Transient , Neumann , grow_rate , A_c , \
                    simtime , verbose_level )

            h_temp_half[0:nx,0:ny] = h_new[0:nx,0:ny]

            # print '1st half ',np.min(h_temp),np.max(h_temp)
            # print '2nd half ',np.min(h_temp_half),np.max(h_temp_half)

            # integration with one full step

            delta_t = delta_t_orig

            h_guess[0:nx,0:ny] = h_temp_half[0:nx,0:ny]

            (h_temp_full[0:nx,0:ny],residual) = \
                    advance_time( h_guess[0:nx,0:ny] , h_old[0:nx,0:ny] , delta_t , delta_x , \
                    delta_y ,  lambda_wb , k_wb , S_c , enne , k , max_nlc , max_inner_iter , res , \
                    Dirichlet, Transient , Neumann , grow_rate , A_c , \
                    simtime ,  verbose_level )

            # print '2nd half ',np.min(h_temp_half),np.max(h_temp_half)
            # print 'full step ',np.min(h_temp_full),np.max(h_temp_full)

            # check the error

            delta_t = delta_t_orig

            delta_h = np.abs( h_temp_half - h_temp_full )

            atol = res
            rtol = res

            err = np.sqrt( nx*ny * np.sum( ( delta_h / ( atol + np.abs(h) * rtol ) )**2 ) )

            if ( verbose_level >= 1 ):

                print('Accuracy = ' + str(err))

            if ( err >= 1.0):

                delta_t_orig = delta_t_orig * 0.975
                if ( verbose_level >= 1 ):
                    print('delta_t_orig =' ,delta_t_orig)

        # err_old = err

        h_init  = h_init + A_c * delta_t

        h_diff = h_temp - h_init

        h[0:nx,0:ny] = h_temp_half[0:nx,0:ny]

        simtime = simtime + delta_t

        print('Time = %.5f delta_t = %.5f err = %10.3E' % (simtime,delta_t,err))

        if ( verbose_level >= 1):

            # gradient is computed using second order accurate central differences in the interior
            # points and either first or second order accurate one-sides (forward or backwards)
            # differences at the boundaries
            h_x,h_y = np.gradient(h,delta_x,delta_y)

            # magnitude of gradient
            grad_h = np.sqrt( h_x**2 + h_y**2 )

            max_angle[iter] = np.arctan(np.max(np.max(grad_h)))*180/np.pi
            mean_angle[iter] = np.arctan(np.mean(grad_h))*180/np.pi

            print('Maximum/mean slope of the DEM = ' + str(max_angle[iter]) \
                  + str(mean_angle[iter]))

        if ( simtime >= iteration_draw_times[iter] ):

            iter += 1
            print('Saving output '+str(iter))

            # add output to netcdf file
            t[iter] = iteration_draw_times[iter-1]
            z[iter,:,:] = h
            dz_dt[iter,:,:] = ( h - h_old ) / delta_t
            dz_tot[iter,:,:] = h - h_init

            if ( plot_output_flag):

                time_text = make_plot(X,Y,h,h_init,h_min,h_max,simtime,cr_angle,run_name, \
                                      iter,plot_show_flag)

            if ( save_output_flag):

                # Save topography on ascii raster file
                header = "ncols     %s\n" % h.shape[1]
                header += "nrows    %s\n" % h.shape[0]
                header += "xllcenter " + str(np.amin(X)) +"\n"
                header += "yllcenter " + str(np.amin(Y)) +"\n"
                header += "cellsize " + str(np.abs(X[1,2]-X[1,1])) +"\n"
                header += "NODATA_value -9999\n"

                output_full = run_name + '_{0:03}'.format(iter) + '.asc'
                print('Saving ascii rater file '+output_full)

                np.savetxt(output_full, np.flipud(h), header=header, fmt='%1.5f',comments='')


    if sys.version_info >= (3, 0):

        elapsed = (time.process_time() - start)

    else:

        elapsed = (time.clock() - start)

    print ('Time elapsed ' + str(elapsed) + ' sec.')
    print ('')

    ncfile.close(); print('Dataset is closed!')


    h_new[0:nx,0:ny] = h[0:nx,0:ny]


    total_area = ( X[0,ny-1] - X[0,0] ) * ( Y[nx-1,0] - Y[0,0] )
    print('Total volume eroded (m^3) = ' + str(np.sum(h_new-h_init)*total_area))

    print('Max. deposition (m) = ' + str(np.maximum(0.0,np.max(h_diff))))

    print('Max. erosion (m) = ' + str(np.maximum(0.0,-np.min(h_diff))))

    return ( h_new , h_diff , time_iter , max_angle , mean_angle )



#!/usr/bin/env python3

import shutil
import os
import numpy as np
from create_synth_landform import create_synth_landform
from ErosioNL import ErosioNL
from read_asc import read_asc

# Script to run the non linear diffusion code on various initial forms
# 2D non linear diffusion code is by
# Mattia de' Michieli Vitturi (Istituto Nazionale di Geofisica e Vulcanologia)
# April 20, 2016


# First, build an initial form


run_name = 'synth_cone'

# Synthetic cinder cone (Hooper & Sheridan):

# computational grid parameters
x_min = -500.0
delta_x = 10.0
x_max = 500.0
y_min = x_min
delta_y = delta_x
y_max = x_max

# parameters defining the cone

# semi-axis of the base of the cinder cone:
# a - semiaxis in the x-direction
# b - semiaxis in the y-direction
a = 400
b = 400

# r is defined as sqrt( (x/a)^2 + (y/b)^2 )
# r = 1 defines the base of the cone
# there are 4 regions defined:
# r<r1      h = h1                                (flat region inside the crater)
# r1<r<r2   h varying linearly between h1 and h2  (crater)
# r2<r<r3   h varying linearly between h2 and h3  (cone slope)
# r<1       h = h3                                (flat ragion outside the cone)

r1 = 0.25 #radius for elevation profile
r2 = 0.5

h1 = 100
h2 = 130
h3 = 0

symmetry = 'radial'

[X, Y, h, k] = create_synth_landform(x_min, delta_x, x_max, y_min, \
                                     delta_y, y_max,r1,r2,h1,h2,h3,a,b,symmetry)

"""
# Linear case:
run_name = 'synth_linear'

# computational grid parameters
x_min = -500.0
delta_x = 10.0
x_max = 500.0
y_min = x_min
delta_y = delta_x
y_max = x_max

# parameters defining the linear profile

# semi-axis of the base of the cinder cone:
# a - position of last slope change
# b - not used
a = 200
b = 200

# r is defined as x/a
# r = 1 defines the end of slope change zone
# there are 4 regions defined:
# r<=r1      h = h1                                (first flat region)
# r1<r<r2    h varying linearly between h1 and h2  (first constant slope region)
# r2<r<r1    h varying linearly between h2 and h3  (second constant slope region)
# r>=1       h = h3                                (final flat ragion)

r1 = 0.25
r2 = 0.5

h1 = 130
h2 = 160
h3 = 0

symmetry = 'y'

[X, Y, h, k] = synth_landform(x_min, delta_x, x_max, y_min, \
                              delta_y, y_max,r1,r2,h1,h2,h3,a,b,symmetry)
"""

# Topography from ascii raster file

"""
run_name = 'SPcrater'

ascii_file = 'sp_crater.asc'
[X, Y, h,x_min,x_max,delta_x,y_min,y_max,delta_y] = read_asc(ascii_file)
"""


# Save initial topography on ascii raster file
header = "ncols     %s\n" % h.shape[1]
header += "nrows    %s\n" % h.shape[0]
header += "xllcenter " + str(x_min) +"\n"
header += "yllcenter " + str(y_min) +"\n"
header += "cellsize " + str(delta_x) +"\n"
header += "NODATA_value -9999\n"

output_full = run_name + '_init.asc'

np.savetxt(output_full, np.flipud(h), header=header, fmt='%1.5f',comments='')
print(output_full+' saved')


# Initialize the mask defining the region to uplift/depression and/or tilt

# no uplift and tilt

nx = h.shape[0]
ny = h.shape[1]


mask = np.zeros((nx,ny))


x0 = 0.0    # x-center of tilt
y0 = 0.0    # y-center of tilt
alfa = 0.0  # angle defining the tilt-axis
c0 = 0.0    # depression in the ragion defined by the mask
c1 = 0.0    # no tilting along the axis parallel to the line alfa=0
c2 = 0.0    # no tilting along the axis orthogonal to the line alfa=0


# Now set up model input
# X, Y, h have been built already

final_time = 1000.0   # final time in kilo years

delta_t_max = 10.00  # maximum time step
delta_t0 = 0.01      # initial time step

cr_angle = 33.0      # critical slope in degrees

enne = 2.0           # exponent for the nonlinearity of the model
                     # enne = np.inf  gives a linear model

#k = 1.0              # m^2/kyr; the diffusion coefficient, a scalar or a (nx,ny) array
                     # the time unit is the same of the parameter 'final_time'

# The diffusion coefficient is multiplied by a relative coefficient k_rel, which
# is defined by a Weibull cumulative distribution with parameters lambda_wb and k_wb
   
lambda_wb = 20       # lambda_wb is the time scale parameter of the Weibull cumulative 
                     #   distribution (for t=lambda_wb, k_rel=(e-1)/e). If lambda_wb=0,
                     #   then k_rel is set to 1.0
                                          
k_wb = 5             # k_wb is the shape parameter of the Weibull distribution;

max_nlc = 10.0       # maximum value of the nonlinear coefficient

max_inner_iter = 100 # the maximum number of iteration for the inner loop.
                     # <20 should be good;

res = 1.e-4          # [m] the residual required for the convergenge of the inner
                     # loop.

bc = 'NNNN'  # the boundary condition : 'N' (Neumann) or 'D' (Dirichlet)
             # or 'T' (Transient). The order is S,W,N,E.
             # Dirichlet is fixed values (elevation in this case)
             # Neumann is fixed gradient (flux null in this case)
             # Transient is elevation at the boundaries changing at
             # fixed rate, given by the parameter 'grow_rates'

gr = 0.0
grow_rates = [ gr, gr, gr, gr]  # rate of change at the boundaries
                             # Used only when the b.c. is 'T'

n_output = 10  # number of output plotted



vx = X - x0
vy = Y - y0

alfarad = np.deg2rad(alfa)

A_c0 = c0 * mask
A_c1 = c1 * mask * ( vx * np.cos(alfarad) + vy * np.sin(alfarad) )
A_c2 = c2 * mask * ( vx * np.sin(alfarad) - vy * np.cos(alfarad) )
A_c = A_c0 + A_c1 + A_c2


verbose_level = 0   # level of output on screen (>1 for debug purposes)
plot_output_flag = 1
save_output_flag = 1
plot_show_flag = 0

#search if another run with the same base name already exists
i = 0

condition = True

base_name = run_name

while condition:

    run_name = base_name + '_{0:03}'.format(i)

    backup_advanced_file = run_name + '_advanced_inp.bak'
    backup_file = run_name + '_inp.bak'

    condition = os.path.isfile(backup_file)

    i = i + 1

# create a backup file of the input parameters
shutil.copy2('run_code_simple.py', backup_file)


( h_new , h_diff , time_iter , max_angle , mean_angle  ) =       \
    ErosioNL(X,Y,h,final_time,delta_t_max,delta_t0,              \
    cr_angle,enne,k,lambda_wb,k_wb,max_nlc,max_inner_iter, res,  \
    bc, grow_rates,run_name , n_output , A_c , verbose_level,    \
    save_output_flag,plot_output_flag,plot_show_flag)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nlerosion",
    version="1.0.0",
    author="Mattia de' Michieli Vitturi",
    author_email="demichie@gmail.com",
    description="A python package for nonlinear erosion modeling",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/demichie/NLerosion",
    packages=setuptools.find_packages(),
    classifiers=(
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
    ),
)



def create_synth_landform(x_min, delta_x, x_max, y_min, delta_y, y_max,r1,r2,h1,h2,h3,a,b,symmetry):

    import numpy as np

    nx = np.int(np.ceil( ( x_max - x_min ) / delta_x ) ) + 1
    ny = np.int(np.ceil( ( y_max - y_min ) / delta_y ) ) + 1

    x = np.linspace(x_min,x_max,nx)
    y = np.linspace(y_min,y_max,ny)


    X,Y = np.meshgrid(x,y)


    h = np.zeros((ny,nx)) # initialization of the array of the altitudes

    for i in range(0,nx):

        for j in range(0,ny):

            if ( symmetry == 'radial' ):

                r = np.sqrt(x[i]**2/a**2+y[j]**2/b**2)

            elif ( symmetry == 'y'):

                r = x[i]/a

            elif ( symmetry == 'x'):

                r = y(j)/b

            else:

                sys.exit("Please define symmetry: radial, x, or y)")

            if ( r < r1 ):

                h[j,i] = h1

            elif ( r < r2 ):

                h[j,i] = h1 + ( r - r1 ) / (r2 - r1) * (h2 - h1)

            elif ( r < 1 ):

                h[j,i] = h2 + ( r - r2 ) / (1.0 - r2) * (h3 - h2)

            else:

                h[j,i] = h3


    # diffusion coefficient

    # there are 5 regions defined:
    # r<r1      k=k1
    # r1<r<r2   k changes linearly between k1 and k2
    # r2<r<r3   k=k2
    # r3<r<r4   k changes linearly between k2 and k1
    # r2>r4     k=k1

    r1 = 0.25 #relative radius
    r2 = 0.4
    r3 = 0.6
    r4 = 0.75

    k1 = 1
    k2 = 1 #one for rim

    k = np.zeros((ny,nx))  # initialization of the array of the erosion coefficients

    for i in range(0,nx):# a loop to define erosion coefficients for all the cells/columns of the grid

        for j in range(0,ny):

            r = np.sqrt(x[i]**2/a**2+y[j]**2/b**2)

            if ( r < r1 ):

                k[j,i] = k1

            elif ( r < r2 ):

                k[j,i] = k1 + ( r - r1 ) / (r2 - r1) * (k2 - k1) # j ( or first index) for the row, i for the columns

            elif ( r < r3 ):

                k[j,i] = k2;

            elif ( r < r4 ):

                k[j,i] = k2 + ( r - r3 ) / (r4 - r3) * (k1 - k2)

            else:

                k[j,i] = k1


    return (X, Y,h,k)



def make_plot(X,Y,Z,Z_init,h_min,h_max,simtime,cr_angle,run_name,iter,plot_show_flag):

    from mpl_toolkits.mplot3d import Axes3D
    from matplotlib import cm
    from matplotlib import colors
    from matplotlib.ticker import LinearLocator, FormatStrFormatter
    import matplotlib.pyplot as plt
    import numpy as np

    import imp

    try:
        imp.find_module('mayavi')
        found = True
    except ImportError:
        found = False
        
    if found:
    
        from mayavi import mlab    

    plt.ion()
    plt.rcParams.update({'font.size':8})

     
    plt.close('all') 
     
    Z_diff = Z-Z_init 
        
    fig = plt.figure()
    fig.set_size_inches(11,7)
    ax1 = fig.add_subplot(2,2,1,projection='3d')
    ax2 = fig.add_subplot(2,2,2)
    ax3 = fig.add_subplot(2,2,3)
    ax4 = fig.add_subplot(2,2,4)
    plt.tight_layout(pad=4, w_pad=4, h_pad=4)
    time_text = ax3.text(0,0, 'time ='+"{:8.2f}".format(simtime)+'s')
    
    delta_x = X[0,1]-X[0,0]
    delta_y = Y[1,0]-Y[0,0]
    
    Z_x,Z_y = np.gradient(Z,delta_x,delta_y)

    grad_Z = np.sqrt( Z_x**2 + Z_y**2 )

    slope = ( np.arctan(grad_Z)*180.0/np.pi )
    
    my_col = cm.jet(slope/cr_angle)

    norm = colors.Normalize(vmin=0.0, vmax=cr_angle)

    ax1.plot_surface(X, Y, Z, rstride=1, cstride=1, facecolors = my_col , 
                    linewidth=0, antialiased=False, alpha = 0.7)
     
    #m = cm.ScalarMappable(cmap=plt.cm.jet, norm=norm)
    #m.set_array([])
    #plt.colorbar(m,ax=ax1,location='bottom') 
                    
    # clb1.set_label('Slope [°]')
                    
                    
    nx = X.shape[1]
    ny = X.shape[0]  
    
    idx1 = int(np.floor(nx/4)) 
    idx2 = int(np.floor(nx/2)) 
    idx3 = int(np.floor(3*nx/4)) 
    
    dh = h_max-h_min
    
    ax1.plot3D(X[idx1,:],Y[idx1,:],Z[idx1,:]+0.01*dh, 'blue')
    ax1.plot3D(X[idx2,:],Y[idx2,:],Z[idx2,:]+0.01*dh, 'red')
    ax1.plot3D(X[idx3,:],Y[idx3,:],Z[idx3,:]+0.01*dh, 'green')

    if ( np.min(Z_diff) == np.max(Z_diff) ):
    
        Z_diff[0,0] = 0.001
        Z_diff[-1,-1] = -0.001
        
   
    ax1.set_xlabel('x [m]')
    ax1.set_xlim(np.amin(X),np.amax(X))
    ax1.set_ylabel('y [m]')
    ax1.set_ylim(np.amin(Y),np.amax(Y))
    ax1.set_zlabel('z [m]')
    ax1.set_zlim(h_min,h_max)

    
    extent = [np.amin(X),np.amax(X),np.amin(Y),np.amax(Y)] 
    
    z_range = np.amax(np.abs(Z_diff))   
    cnt = ax2.imshow(Z_diff, extent=extent, vmin=-z_range, vmax=z_range)
    
    clb = plt.colorbar(cnt,ax=ax2)
    clb.set_label('Delta h [m]')
    # colorbar(cnt);
    
    ax2.set_xlabel('x [m]')
    ax2.set_xlim(np.amin(X),np.amax(X))
    ax2.set_ylabel('y [m]')
    ax2.set_ylim(np.amin(Y),np.amax(Y))
    

    
    l1, = ax3.plot(X[idx1,:],Z[idx1,:], 'b-')
    l2, = ax3.plot(X[idx2,:],Z[idx2,:], 'r-')
    l3, = ax3.plot(X[idx3,:],Z[idx3,:], 'g-')
    
    ax3.legend((l1, l2, l3), ("y="+str(Y[idx1,0])+' m',"y="+str(Y[idx2,0])+' m', \
                             "y="+str(Y[idx3,0])+' m'), loc='upper right', shadow=True)

    
    ax3.plot(X[idx1,:],Z_init[idx1,:],'b--')
    ax3.plot(X[idx2,:],Z_init[idx2,:],'r--')
    ax3.plot(X[idx3,:],Z_init[idx3,:],'g--')
           
    ax3.set_xlabel('x [m]')      
    ax3.set_ylabel('z [m]')    
    
    x_min, x_max = ax3.get_xlim()
        
    y_min, y_max = ax3.get_ylim()
    
    time_text.set_position((x_min+0.05*(x_max-x_min), y_min+0.9*(y_max-y_min)))
    time_text.set_text('time ='+"{:8.2f}".format(simtime)+'s')

 
    slope_flat = slope.flatten()
    
    """
    n, bins, patches = ax4.hist(slope_flat, weights=100.0*np.ones(len(slope_flat)) / len(slope_flat),\
             histtype='stepfilled', alpha=0.2)
             
    bin_centers = 0.5 * (bins[:-1] + bins[1:])

    # scale values to interval [0,1]
    col = bin_centers - min(bin_centers)
    col /= max(col)

    cm = plt.cm.get_cmap('RdYlBu_r')
    
    for c, p in zip(col, patches):
        plt.setp(p, 'facecolor', cm(c))   
    """
        
    # Get the histogramp
    cm = plt.cm.get_cmap('jet')

    Yh,Xh = np.histogram(slope_flat)
    x_span = Xh.max()-Xh.min()
    C = [cm(x/cr_angle) for x in Xh]

    ax4.bar(Xh[:-1]+0.5*(Xh[1]-Xh[0]),Yh,color=C,width=Xh[1]-Xh[0])      

    ax4b = ax4.twinx()
    
    # plot the cumulative histogram
    n_bins = 50
    ax4b.hist(slope_flat, n_bins, density=True, histtype='step',
                           cumulative=True, label='Empirical')

    ax4.set_xlabel('slope [degrees]')
    ax4.set_ylabel('frequency [%]')
    ax4b.set_ylabel('cumulative distribution')

    frame_name = run_name + '_{0:03}'.format(iter) + '.png'
    plt.savefig(frame_name,dpi=200)

    frame_name = run_name + '_{0:03}'.format(iter) + '.pdf'
    plt.savefig(frame_name)

    if plot_show_flag:

        plt.show()
        plt.pause(0.01)
    
    return time_text
